#!/usr/bin/env python3
"""
Torrent Archiver

Provides functionality for managing datasets distributed through BitTorrent.
It tracks files and reconciles hardlinks between download directories and
archival locations.

"""

import os
import sys
import re
import uuid
import argparse
from datetime import datetime, timezone

import qbittorrentapi
from sqlalchemy import create_engine, inspect
from sqlalchemy.orm import Session

from .models import Base, SchemaVersion, Client

# SCHEMA format is YYYYMMDDX
SCHEMA = 202503100


def init_db(engine):
    """
    Initialize database
    """
    Base.metadata.create_all(engine)

    with Session(engine) as session:
        if not session.query(SchemaVersion).first():
            now = datetime.now(timezone.utc)
            version = SchemaVersion(version=SCHEMA, applied_at=now)
            session.add(version)
            session.commit()


def get_schema_version(engine):
    """
    Get current schema version from database
    """
    with Session(engine) as session:
        version = session.query(SchemaVersion).order_by(SchemaVersion.id.desc()).first()
        return version.version if version else None


def list_tables(engine):
    """
    List all tables in database
    """
    inspector = inspect(engine)
    return inspector.get_table_names()


def add_client(engine, name, endpoint, last_seen):
    """
    Add a new client endpoint to database
    """
    with Session(engine) as session:
        client = Client(
            uuid=str(uuid.uuid4()), name=name, endpoint=endpoint, last_seen=last_seen
        )
        session.add(client)
        session.commit()


def find_client(engine, endpoint):
    """
    Find existing client
    """
    with Session(engine) as session:
        clients = (
            session.query(Client.id, Client.name, Client.uuid)
            .filter_by(endpoint=endpoint)
            .all()
        )
        return clients


def list_clients(engine):
    """
    List all stored clients
    """
    with Session(engine) as session:
        return session.query(Client).all()


def auth_qbittorrent(endpoint, username, password):
    """
    Authenticate with the qBittorrent client
    """
    qb = qbittorrentapi.Client(host=endpoint, username=username, password=password)
    try:
        qb.auth_log_in()
    except qbittorrentapi.LoginFailed as e:
        raise ValueError(f'Login failed for endpoint "{endpoint}"') from e
    except Exception as e:
        raise ValueError(f"An unexpected error occurred: {str(e)}") from e
    if not re.match(r"^v?\d+(\.\d+)*$", qb.app.version):
        raise ValueError(f'Invalid version "{qb.app.version}" found at "{endpoint}"')
    return qb


def scan_torrents(qb_client, debug=False):
    """
    Scan torrents using the provided qBittorrent client.
    """
    torrents = qb_client.torrents_info()
    print(f"[INFO]: There are {len(torrents)} torrents\n")
    for torrent in torrents[:2]:
        files = qb_client.torrents_files(torrent.hash)
        trackers = qb_client.torrents_trackers(torrent.hash)
        print(f"[name]: {torrent.name}")
        print(f"[infohash_v1]: {torrent.hash}")
        print(f"[content_path]: {torrent.content_path}")
        print(f"[magnet_uri]: {torrent.magnet_uri[:80]}")
        print(f"[completed_on]: {torrent.completed}\n")
        print(f"[trackers]: {len(trackers)}")
        print(f"[file_count]: {len(files)}\n")
        if debug:
            print(f"[DEBUG]: {repr(torrent)}")
            for elem in trackers:
                print(f"[DEBUG]: Tracker {repr(elem)}")
            print("\n", end="")


def scan(args, engine):
    """
    Scan command to authenticate with the qBittorrent client and scan torrents.
    """
    if args.name:
        clients = find_client(engine, args.name)
        if len(clients) == 1:
            client_info = clients[0]
            qb_client = auth_qbittorrent(client_info.endpoint, client_info.username, args.password)
            scan_torrents(qb_client, debug=args.debug)
        elif len(clients) == 0:
            raise ValueError(
                f'Client with name "{args.name}" not found. '
                "Please use the 'client add' command to add a new client."
            )
        else:
            raise ValueError(
                f"Multiple clients with the same name: {args.name}"
            )
    elif args.directory:
        print("[INFO]: --directory is not implemented")
    else:
        raise ValueError("Must specify directory OR client name")


def client_add(args, engine):
    """
    Add a new client to the database.
    """
    if args.name and args.endpoint and args.username:
        now = datetime.now(timezone.utc)
        add_client(engine, args.name, args.endpoint, now)
        print(f"[INFO]: Added client {args.name} ({args.endpoint})")
    else:
        raise ValueError("Must specify --name, --endpoint, and --username to add a client")


def client_list(engine):
    """
    List all stored clients.
    """
    clients = list_clients(engine)
    for client in clients:
        print(f"{client.name} ({client.endpoint}) - Last seen: {client.last_seen}")


def main():
    """
    Parses command-line arguments and executes the corresponding command.
    """
    parser = argparse.ArgumentParser(description="Manage BT archives", prog="tarc")
    subparsers = parser.add_subparsers(
        dest="command", required=True, help="Available commands"
    )

    # scan command
    scan_parser = subparsers.add_parser("scan", help="Scan torrents")
    scan_parser.add_argument("--debug", action="store_true", help="Enable debug mode")
    scan_parser.add_argument("-n", "--name", help="Name of client")
    scan_parser.add_argument("-d", "--directory", help="Directory to scan")
    scan_parser.add_argument("-u", "--username", help="Username")
    scan_parser.add_argument("-p", "--password", help="Password")

    # client command
    client_parser = subparsers.add_parser("client", help="Manage clients")
    client_subparsers = client_parser.add_subparsers(dest="client_command", required=True)

    # client add command
    client_add_parser = client_subparsers.add_parser("add", help="Add a new client")
    client_add_parser.add_argument("-n", "--name", required=True, help="Name of client")
    client_add_parser.add_argument(
        "-e", "--endpoint", required=True, help="Endpoint URL"
    )

    # client list command
    client_list_parser = client_subparsers.add_parser("list", help="List all clients")

    args = parser.parse_args()

    # Check for valid subcommand for client
    if args.command == "client" and args.client_command is None:
        parser.error("The 'client' command requires a subcommand (add or list).")

    storage_path = os.path.expanduser("~/.tarc.db")
    engine = create_engine(f"sqlite:///{storage_path}")

    if not list_tables(engine):
        print(f"[INFO]: Initializing database at {storage_path}")
        init_db(engine)

    schema_found = get_schema_version(engine)
    if schema_found != SCHEMA:
        raise ValueError(f"SCHEMA {schema_found}, expected {SCHEMA}")

    try:
        if args.command == "scan":
            scan(args, engine)
        elif args.command == "client" and args.client_command == "add":
            client_add(args, engine)
        elif args.command == "client" and args.client_command == "list":
            client_list(engine)
    except ValueError as e:
        print(f"[ERROR]: {e}")
        sys.exit(1)


if __name__ == "__main__":
    main()
